/* XML_lex.C -- tokenize XML/HTML files. */
/***********************************************************
 *                                                         *
 *   (C) Copyright 2000-2002, Burton Systems Software.    *
 *                   All Rights Reserved.                  *
 *                                                         *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                         *
 *  OEMs should NOT configure TLIB to match this keyflag:  *
 *+=>keyflag<=-- "&(#)%n, Burton version %v, %d "         *
 *  "&(#)XML_lex.c, Burton version 42, 03-Apr-02 "
 ***********************************************************/

/*
   build with VC++ 5.0 or 6.0 under Win32 OSs like this:
     cl XML_lex.c
   or build w/ C6 under OS/2 like this (untested):
     cl -c XML_lex.c & link XML_lex.obj; & bind XML_lex.exe
*/

#if (defined _WIN32) && (!defined WIN32)
  #define WIN32
#endif
#if (defined WIN32)
  #define FLAT32 1
#else
  #define FLAT32 0
#endif
#define STRICT
#ifdef WIN32
  #define WIN32_LEAN_AND_MEAN  /* this makes compiles faster, but it is also necessary to avoid redefinition of typedef boolean */
  #include <windows.h>  /* define FindFirstFile() and FindClose() */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#define LIBPTR_TO_STDPTR
#ifndef _FAR_
  #define _FAR_
#endif
typedef enum {false=0, true=1} boolean;
typedef boolean (* callback_t)(char *,int);
int InpEOF;

#if FLAT32
  #define far
  #define _far
  #define _fmalloc malloc
  #define _nmalloc malloc
  #define _ffree free
  #define _nfree free
  #define _fmemmove memmove
  #define _fmemcpy memcpy
  #define _fmemset memset
#endif

int debugmode = 0;  /* -d */
int showlinenums = 0;  /* -n */
int verboseoutput = 0;  /* -v */

char * infile;
char * outfile;
char infilename[264];
char outfilename[264];


void my_fclose( FILE _FAR_ * fil )
  {
   int handle;
    if (fil)
      {
        if (debugmode)
          {
#ifdef _WINDOWS
            handle = _fileno( fil );
#else
            handle = fileno( fil );
#endif
            printf( "my_fclose(%08lX) h=%d\n", (FILE _far *)fil, handle );
          }
        fclose( fil );
      }
  } /* my_fclose */

char *stristr(const char *String, const char *Pattern)
{
      char *pptr, *sptr, *start;
      unsigned int  slen, plen;
      for (start = (char *)String,
           pptr  = (char *)Pattern,
           slen  = strlen(String),
           plen  = strlen(Pattern);
           /* while string length not shorter than pattern length */
           slen >= plen;
           start++, slen--)
      {
            /* find start of pattern in string */
            while (toupper(*start) != toupper(*Pattern))
            {
                  start++;
                  slen--;
                  /* if pattern longer than string */
                  if (slen < plen)
                        return(NULL);
            }
            sptr = start;
            pptr = (char *)Pattern;
            while (toupper(*sptr) == toupper(*pptr))
            {
                  sptr++;
                  pptr++;
                  /* if end of pattern then pattern was found */
                  if ('\0' == *pptr)
                        return (start);
            }
      }
      return(NULL);
}
#if FLAT32
  #define _fstristr stristr
#endif

/* Test whether a specified path is "rooted," that is, whether it begins
   with "d:\" or "\".  */
static boolean rooted_path( const char * fname )
  {
    return   (    ('\\'==fname[0])
               || (fname[0] && (':'==fname[1]) && ('\\'==fname[2]))
             );
  } /* rooted_path */

/* Return subscript in str[] of last occurance of ch.  Return -1 if not found */
static int lastposC(char ch, const char * s)
  {
   int i, len;
    len = strlen(s);
    for (i = len-1; i >= 0; i--)
        if (s[i] == ch)
            return i;
    return -1;
  } /* lastposC */

/* Return position of the last backslash or slash or colon in a string.
   Returns -1 for failure, or the index of the character for success (e.g.,
   if the first character in the string is a backslash, it returns "0").
   For a function which returns 0 for failure, call LastSlashOrColonP().  */
static int LastSlashOrColonC(const char * s)
  {
   int i, j;
    i = lastposC(':',s);
    j = lastposC('\\',s);
    if (j > i)
        i = j;
    j = lastposC('/',s);
    if (j > i)
        i = j;
    return i;
  } /* LastSlashOrColonC */

/* delete "subdir\.." or ".\" from path */
void remove_doubledots( char * path )
  {
   char * tmp1, * tmp2;
    /* first change all "\.\"s to "\"s */
    while ((path[0]=='.') && (path[1]=='\\'))
      { /* path starts with ".\" -- delete it */
        strcpy( path, path+2 );
      }
    while (NULL != (tmp2=LIBPTR_TO_STDPTR strstr(path,"\\.\\")))
      { /* path contains a "\.\" */
        strcpy( tmp2, tmp2+2 );
      }
    /* then change all "\subdir\..\"s to "\"s */
    while (NULL != (tmp2=LIBPTR_TO_STDPTR strstr(path,"\\..\\")))
      { /* path contains a ".." */
        *tmp2 = (char)'\0';  /* so strrchr finds last '\' before the '\..\' */
        tmp1 = LIBPTR_TO_STDPTR strrchr( path, '\\' );
        *tmp2 = (char)'\\';  /* restore */
        if (!tmp1)
            break;  /* there's no '\' before the "\\..\\", so leave it alone */
        if (1==(tmp2-tmp1))
            break;  /* double slash not understood, so leave it alone */
        if ('.'==tmp1[1])
            break;  /* subdir started with '.', very weird, so leave it alone */
        strcpy( tmp1, tmp2+3 );
      }
  } /* remove_doubledots */


/* Similar to strncpy (string copy) except that we only copy as many bytes
   as necessary, and dest[] is certain to be 0-terminated because we stuff
   a 0-byte after the last character.  Note that this means that dest[]
   *MUST* be at least maxlen+1 bytes long!  If (maxlen<=0) or (src==NULL),
   then only the '\0' terminator is stored.  However, if dest is NULL, then
   this function is a NO-OP. */
static void my_strncpy(char *dest, const char *src, int maxlen)
  {
   register int len;
    if (dest)
      {
        len = 0;
        if ((maxlen > 0) && src)
           len = strlen(src);
        if (!len)
            dest[0] = '\0';
        else if (len < maxlen)
            strcpy( dest, src );
        else
          {
            strncpy( dest, src, (size_t)maxlen );
            dest[maxlen] = '\0';
          }
      }/*if*/
  } /* my_strncpy */

/* Like strncat(), but ensures that dest is null-terminated when done, and
   the third parameter is a maximum final length instead of a maximum number
   of bytes to copy.
   Note that this means that dest[] *MUST* be at least maxlen+1 bytes long!
   If maxlen == 0, then only the '\0' terminator is stored.  If src is NULL
   or of zero length, then this function does nothing except store a 0-byte
   at dest[maxlen] to ensure that the result is no longer than maxlen bytes.
   However, if dest is NULL, then this function is a NO-OP. */
static void my_strncat(char *dest, const char *src, int maxlen)
  {
   int len;
    if (dest)
      {
       len = strlen(dest);
       if (len >= maxlen)
          dest[maxlen] = '\0';
       else
          my_strncpy(dest+len, src, maxlen-len);
     }
  } /* my_strncat */

#if 0  /* not used in this version */
/* Similar to fgets(), except that the trailing '\n' is removed, and also
   as a side-effect InpEOF is set to true for end-file, or false otherwise.
   Returns strlen(line), which never exceeds n-1.  */
int my_fgets( char * line, int n, FILE _FAR_ * fil )
  {
   int len;
    line[0] = '\0';
    InpEOF = (NULL==fgets(line,n,fil));
    len=strlen(line);
    while (len && (('\n'==line[len-1]) || ('\r'==line[len-1])))
        line[--len] = '\0';
    /* printf("dbg: my_getfs()=%d, line=\"%s\"\n", len, line ); */
    return len;
  } /* my_fgets */
#endif


/* remove all quote characters from a string (used for filenames & paths) */
void remove_quotes_from_path( char * fname )
  {
   char * qpos;
    for (;;)
      {
        qpos = LIBPTR_TO_STDPTR strchr( fname, '"' );
        if (!qpos)
            break;
        strcpy( qpos, qpos+1 );
      }/*for*/
  } /* remove_quotes_from_path */

/* scaffolded version doesn't do very much */
boolean quick_canonical_path( char * fname, int dbgnum )
  {
    remove_quotes_from_path( fname );
    remove_doubledots( fname );
    return true;
  }

/* Return pointer to the first character of the file name; that is, the
   name and extension with the directory part removed. */
/*lint -e605*/
char * name_and_ext( const char * path )
  {
#if (defined _MSC_VER) && (_MSC_VER >= 700)
    #pragma warning(disable:4090)
#endif
    return path + LastSlashOrColonC(path) + 1;
    /* "warning C4090: different 'const/volatile' qualifiers" is normal here */
#if (defined _MSC_VER) && (_MSC_VER >= 700)
    #pragma warning(default:4090)
#endif
  } /* name_and_ext */
/*lint +e605*/


#if 0
#ifdef WIN32
/* Used for a quick determination of the binary length of a file;
   returns -1L for error. */
static long FileLengthViaFindFIrst( char * filespec )
  {
  WIN32_FIND_DATA find_data;
  HANDLE hand;
   hand = FindFirstFile( filespec, &find_data );
   if (INVALID_HANDLE_VALUE == hand)
       return -1L;  /* not found */
   FindClose( hand );
   return find_data.nFileSizeLow;  /* found it */
  } /* FileLengthViaFindFIrst */
#endif
#endif


/* Compare keywd[] to the first strlen(keywd) characters at *bufptr, and
   return 0 if they are the same, non-0 if they are different.  This is a
   case-insensitive compare.  */
static int strcasecmp( char * bufptr, char * keywd )
  {
   int len = strlen(keywd);
    if (!bufptr)
        return !keywd;  /* true if bufptr=NULL=keywd, else false */
    return strnicmp( bufptr, keywd, len );
  }


enum { not_in_tok_c=0, in_tok_c=1 } lex_state = not_in_tok_c;

int XML_tokn_is_tag = false;  /* 1 if XML tag other than comment, 0 if text or <!--comment--> */
int XML_tokn_is_truncated = false;  /* 1 if more to come */
int XML_tokn_is_continuation = false;  /* 1 if previous result was truncated */
int XML_tokn_is_comment = false;  /* 1 if a comment tag, "<!-- ... -->" */
long int XML_tokn_line_no = 1L;
long int XML_nxt_line_no = 1L;


/* we need a stream-like file structure that permits us to peek ahead and
   to unget more than one character */
#define AUG_BLKSIZ 128
#define AUG_MAX_UNGETC 20
/* augmented file */
typedef struct {
      FILE _FAR_ * fn;
      int nxt, len;
      char auxbuf[ AUG_BLKSIZ + AUG_MAX_UNGETC + 2 ];
    } augFILE_t;


/* File stream should already be open; convert to an "augmented stream"
   (which involves a _fmalloc), and initialize the XML parse state. */
augFILE_t _FAR_ * init_XML_parse( FILE _FAR_ * fil )
  {
   augFILE_t _FAR_ * result;
    lex_state = not_in_tok_c;
    XML_tokn_is_tag = false;  /* 1 if XML tag other than comment; else 0 */
    XML_tokn_is_truncated = false;  /* 1 if more to come */
    XML_tokn_is_continuation = false;  /* 1 if previous result was truncated */
    XML_tokn_is_comment = false;  /* 1 if a comment tag, "<!-- ... -->" */
    XML_nxt_line_no = XML_tokn_line_no = 1L;
    result = malloc( sizeof(augFILE_t) );
    if (result)
      {
        _fmemset( result, 0, sizeof(augFILE_t) );
        result->fn = fil;
        result->nxt = result->len = AUG_MAX_UNGETC;  /* no chars yet read */
      }
    else
        my_fclose( fil );  /* out of memory */
    return result;
  } /* init_XML_parse */


/* Close file stream and free the buffer.  To prevent the file close,
   set af->fn to 0 before calling this function. */
void uninit_XML_parse( augFILE_t _FAR_ * af )
  {
    lex_state = not_in_tok_c;
    XML_tokn_is_tag = false;  /* 1 if XML tag other than comment; else 0 */
    XML_tokn_is_truncated = false;  /* 1 if more to come */
    XML_tokn_is_continuation = false;  /* 1 if previous result was truncated */
    XML_tokn_is_comment = false;  /* 1 if a comment tag, "<!-- ... -->" */
    if (af)
      {
        if (af->fn)
            my_fclose( af->fn );
        _fmemset( af, 0, sizeof(augFILE_t) );  /* not really necessary, of course */
        free( af );
      }
  } /* uninit_XML_parse */


/* Read some more of the file, so that (unless we hit EOF) there will be
   at least AUG_BLKSIZ additional unread bytes that are already in the
   buffer.  Usually there will be exactly AUG_BLKSIZ additional unread
   bytes, unless at EOF, but if aug_ungetc has been called then there
   might be more than AUG_BLKSIZ additional unread bytes.  */
void aug_fillbuf( augFILE_t _FAR_ * af )
  {
   int ulen;
    if (af->nxt < af->len)
      { /* there are still some unread characters in auxbuf */
        ulen = af->len - af->nxt;
        if (ulen >= AUG_BLKSIZ)
            return;  /* auxbuf is already full */
        memmove( af->auxbuf + AUG_MAX_UNGETC, af->auxbuf + af->nxt, ulen );
        af->nxt = AUG_MAX_UNGETC;
        af->len = ulen + AUG_MAX_UNGETC;
      }
    else
        af->nxt = af->len = AUG_MAX_UNGETC;
    ulen = fread( af->auxbuf + af->len, 1, AUG_BLKSIZ + AUG_MAX_UNGETC - af->len, af->fn );
    af->len += ulen;
    if (af->len < (AUG_BLKSIZ+AUG_MAX_UNGETC))
      { /* fill the rest of the buffer with zero bytes */
        _fmemset( af->auxbuf + af->len, 0, (AUG_BLKSIZ+AUG_MAX_UNGETC) - af->len );
      }
  } /* aug_fillbuf */


/* augmented-functionality getc() */
int aug_getc( augFILE_t _FAR_ * af )
  {
    if (af->nxt >= af->len)
      {
        aug_fillbuf( af );
        if (af->nxt >= af->len)
            return EOF;  /* -1 */
      }
    return (int)(af->auxbuf[af->nxt++]);
  } /* aug_getc */


/* Unlike the standard ungetc() function, this one can be used to put back
   more than one character: up to AUG_MAX_UNGETC characters. */
void aug_ungetc( int c, augFILE_t _FAR_ * af )
  {
    if (af->nxt)
        af->auxbuf[--(af->nxt)] = (char)c;
  } /* aug_ungetc */


#if 0  /* not used in this version, but might need it in a future version */
/* Pass in the number of unread characters that you'd like to be able to peek,
   and this function will return a pointer to at least that many of them (zero-
   filled if near EOF).  amt_ahead must not exceed AUG_BLKSIZ. */
char _FAR_ * aug_peek_ahead( int amt_ahead, augFILE_t _FAR_ * af )
  {
   int ulen = af->len - af->nxt;
    if (ulen < amt_ahead)
        aug_fillbuf( af );
    return af->auxbuf + af->nxt;
  } /* aug_peek_ahead */
#endif


/*                  Lexical analysis for XML/HTML files.
      This is like fgets(), but it'll also break at start/end of XML tags,
   and it translates CR or CR+LF to LF alone.  The trailing LF is included
   in the result.
     The result in line_or_tag[] will be at most n-1 characters long, plus
   a terminating '\0'.  Excess characters are returned in the next result.
     Function return value is the token or line length, strlen(line_or_tag),
   which is always in the range 0..n-1.
     XML tags start with '<' and end with '>'.  But if one is longer than
   n-1 characters, it'll be split, so just looking at the first or last
   character of the result is not the surest way to tell that an XML tag
   is being returned.  However, you can always tell that an XML tag (or
   part of a tag) has been returned by checking the global side-effect
   result, XML_tokn_is_tag.
     XML comments are returned as a special case of tags, in which the
   global side-effect result flag XML_tokn_is_comment is true.
     If a text line or XML tag is split, then XML_tokn_is_truncated will be
   set for all except the last fragment, and XML_tokn_is_continuation will
   be set for all except the first fragment.
     If XML_tokn_is_truncated is true, then the result length will usually
   be n-1, except for XML comments, which can be a shorter.
*/
int XML_fgets( char * line_or_tag, int n, augFILE_t _FAR_ * fil )
  {
   int len, lim, soft_lim1, soft_lim2, ch;
   int tag_delim;  /* '<' or '>' */

    len = 0;
    XML_tokn_line_no = XML_nxt_line_no;  /* starting line no. for next token */
    if ((n <= 1) || !line_or_tag)
      {
        if (line_or_tag)
            line_or_tag[0] = '\0';
        XML_tokn_is_truncated = true;
        XML_tokn_is_continuation = false;
        XML_tokn_is_tag = false;
        return len;
      }
    lim = n-1;  /* number of chars that'll fit, not including trailing '\0' */
    soft_lim1 = (2*n)/3;  /* if token exceeds soft_lim2, we try to break it at whitespace, except within quotes */
    soft_lim2 = (6*n)/7;  /* if token exceeds soft_lim, we try to break it at whitespace, even within quotes */

  get_another:
    XML_tokn_is_tag = lex_state;  /* 1 if within token (or comment, for now), else 0 */
    XML_tokn_is_continuation = XML_tokn_is_truncated;  /* if the previously returned line_or_tag was truncated, then this next one is the continuation */
    if (!XML_tokn_is_continuation)
        XML_tokn_is_comment = false;  /* otherwise, it isn't a comment unless we see "<!--" at the beginning (which we've not seen yet) */
    XML_tokn_is_truncated = false;
    if (in_tok_c==lex_state) /* in_tok_c=1 */
        tag_delim = '>';  /* looking for end of tag or comment */
    else
        tag_delim = '<';  /* looking for start of tag or comment */

    for (;;) {
       ch = aug_getc( fil );
       if ((ch == EOF) || (ch == '\r') || (ch == '\n') || (ch == tag_delim))
         {
           if (ch == '\r')
             { /* look ahead one char, checking for CR+LF */
               ch = aug_getc( fil );
               if (ch != '\n')
                   aug_ungetc( ch, fil );  /* not CR+LF, so put back next char */
               ch = '\n';  /* translate CR or CR+LF to LF alone */
             }
           if (ch == '\n')
               XML_nxt_line_no++;
           if ((in_tok_c==lex_state) && (ch == '\n') && !XML_tokn_is_comment)
             { /* within XML token, translate CR, or LF, or CR+LF to a blank,
                  except within a comment or when nearing buffer overflow */
               if (len < soft_lim1)
                   ch = ' ';
               else
                 { /* within a token, if approaching a full buffer we break
                      the token at a newline */
                   XML_tokn_is_truncated = true;
                   break;
                 }
             }
           else if (('>'==ch) && XML_tokn_is_comment)
             { /* perhaps end of comment token */
               int dpos = len-1;  /* likely subscript of 2nd '-' character */
               /* handle the weird but legal case in which there is
                  whitespace between the '--' and the '>' */
               while ((dpos > 1) && isspace(line_or_tag[dpos]))
                   dpos--;
               /* check for '--' preceeding the closing '>' */
               if (   (dpos >= 1)
                   && ('-' == line_or_tag[dpos])
                   && ('-' == line_or_tag[dpos-1]) )
                 {
                   break;  /* found end of a comment token */
                 }
               /* else not end of token, so go ahead and store the '>' as
                  just another character within the comment token */
             }
           else
             { /* end of token or line */
               char * ssi_ptr, * ssi_endptr;
               /* If lexing an XML comment, stop at '-->' (above) or
                  end-of-line (here), but if at end-of-line it is
                  considered to be a truncated token.
                  If not lexing an XML token, stop at end-of-line.
                  If lexing a non-comment XML token, stop at '>'.  */
               if (XML_tokn_is_comment)
                 {
                   XML_tokn_is_truncated = true;  /* at newline, not '-->' */
                   break;  /* end of comment line; more to come */
                 }

               /* Next, a special check for SSI directive embedded within tag */
               /* else  (in_tok_c==lex_state) */
               if (XML_tokn_is_continuation)
                   ssi_ptr = line_or_tag;
               else
                   ssi_ptr = line_or_tag + 1;  /* don't look at 1st '<' */
               line_or_tag[len] = '\0';
               do {
                  /* is there a start-of-SSI-directive in the tag? */
                  ssi_endptr = NULL;
                  ssi_ptr = LIBPTR_TO_STDPTR strstr( ssi_ptr, "<!--#" );
                  if (ssi_ptr)
                    { /* Found the start of an embedded SSI directive within the tag! */
                      /* Question we have to answer is: have we already seen its end?
                         We've just seen (and not yet stored) a '>' character.  So it
                         might be that it part of the "-->" at the end of the SSI
                         directive, rather than being the end of the token.  */
                      ssi_endptr = LIBPTR_TO_STDPTR strstr( ssi_ptr+5, "-->" );
                      if (ssi_endptr)
                        { /* end of SSI directive was already seen, so this '>' is
                             probably the end of the whole token... unless there
                             is a 2nd embedded SSI directive. */
                          ssi_ptr = ssi_endptr+3;
                          /* set ssi_ptr for start of scan for a 2nd SSI directive */
                        }
                      else
                        { /* SSI end directive not yet seen; is this it? */
                          ssi_endptr = line_or_tag + len - 2;
                          if (('-' != ssi_endptr[0]) || ('-' != ssi_endptr[1]))
                              ssi_endptr = NULL;
                          /* If ch is the '>' of a "-->" which is the end of an
                             SSI directive, then ssi_endptr points to "--".  */
                          ssi_ptr = NULL;  /* we're done, break out of the loop */
                        }
                    }
                  /* else  there's no embedded SSI directive, so we're
                           definitely at end-of-token; fall thru with
                           ssi_endptr=ssi_ptr=NULL
                  */
               } while (ssi_ptr);
               if (!ssi_endptr)
                   break; /* not end of SSI directive, so must be end of token */
               /* fall thru if not end of token */
             }
         }/*if*/
       /* Next is a special check to make sure that we don't split an
          end-of-comment delimiter "-->" due to a buffer full condition.
          This isn't quite correct, because it doesn't properly handle
          the (deprecated) case in which 3 or more hyphens are used, nor
          the (permitted but weird) case in which whitespace is placed
          between the '--' and the '>' character.  Algorithm is simply
          to check for a first hyphen (a hyphen that follows a non-hyphen)
          encountered when there is room for only 1 or 2 more characters
          in the line_or_tag[] buffer.  */
       if (    ('-'==ch)
            && ((len==(lim-2)) || (len==(lim-1)))
            && (line_or_tag[len-1] != '-')
            && XML_tokn_is_comment )
         {
           aug_ungetc( ch, fil );  /* put back the '-' */
           ch = line_or_tag[len];  /* previous character */
           XML_tokn_is_truncated = true;
           break;
         }
       /* try to break big tokens at whitespace if len near limit */
       if ((len > soft_lim1) && ((' '==ch) || ('\t'==ch) || ('\014'==ch) || ('\n'==ch)))
         { /* blank, tab, or form-feed */
          int count_quote_marks, ndx;
           if (len > soft_lim2)
             { /* very near to filling output buffer, so break at any
                  whitespace, even if within quote marks */
               if (in_tok_c==lex_state)
                   XML_tokn_is_truncated = true;
               break;
             }
           /* if number of characters is between soft_lim1 & soft_lim2, then
              check whether we've seen an even or odd number of quote marks */
           count_quote_marks = 0;
           for (ndx=0; ndx<len; ndx++)
             {
               if ('"'==line_or_tag[ndx])
                   count_quote_marks++;
             }
           if (0 == (1 & count_quote_marks))
             { /* even number of quote marks, so break at this whitespace */
               if (in_tok_c==lex_state)
                   XML_tokn_is_truncated = true;
               break;
             }
           /* within quoted string, we won't break a token until soft_lim2 */
         }
       /* store the character & count it */
       line_or_tag[len++] = (char)ch;
       /* check for tokens that start with "<!--" because they are comments */
       if (XML_tokn_is_tag && (!XML_tokn_is_continuation) && (4==len))
         {
           XML_tokn_is_comment = !strncmp( "<!--", line_or_tag, 4 );
         }
       /* check for buffer-full condition */
       if (len == lim)
         { /* line_or_tag[] is full! */
           ch = '\0';  /* we already stored it, so set ch=0 so we won't store it again */
           XML_tokn_is_truncated = true;
           break;
         }
    }/*for*/

    /* printf("dbg: final char = %d\n", ch ); */

    if ((ch == '<') && (in_tok_c != lex_state))
      { /* start of an XML tag; don't include '<' in returned result */
        lex_state = in_tok_c;  /* initial state for next time we're called */
        aug_ungetc( ch, fil );  /* put back '<' for next time */
      }
    else if ((EOF != ch) && (len != lim) && ('\0' != ch))
      {
        line_or_tag[len++] = (char)ch;
        /* ch is usually LF or '>', unless XML_tokn_is_truncated is true */
        if ((ch == '>') && !XML_tokn_is_truncated)
          { /* at end of an XML tag */
            lex_state = not_in_tok_c;
          }
      }

    InpEOF = ((EOF == ch) && (0 == len));
    line_or_tag[len] = '\0';

    if (debugmode & 32)
        printf( "XML_gets: l=%d s=\"%s\"\n", len, line_or_tag );

    if ((0==len) && (!XML_tokn_is_tag)
                 && (!XML_tokn_is_continuation) && !InpEOF)
      { /* XML token start ('<') seen at start of line, or right after end
           of another XML token; so don't bother to return 0 bytes of text
           before the upcoming XML token.
        */
        goto get_another;
      }
    /* within the guts of XML_fgets(), it is convenient to have XML_tokn_is_tag
       be true for both regular tokens and comments, but when we return it
       should be false for <!--comments--> even though they're similar to tags. */
    if (XML_tokn_is_tag && XML_tokn_is_comment)
        XML_tokn_is_tag = false;
    return len;
  } /* XML_fgets */


/* Input is the name of an XML or HTML file to read.  Output is the tokens
   in easy-to-read format.  */
void lex1file( char * inpfname, char * outfname )
  {
   FILE _FAR_ * fn, * outf;
   augFILE_t _FAR_ * af;
#if FLAT32
  #define MAX_TOKN_LEN 4096  /* up to 4095 chars/tag or line in 32-bit versions */
#else
  #define MAX_TOKN_LEN 512  /* up to 511 chars/tag or line in 16-bit versions */
#endif
   char token[MAX_TOKN_LEN+2];
   int len;
   long tokn_count = 0L;
   char lprefix[6];
   int i;
   boolean warned = false, warned2 = false;
   int handle;

    if (!strcmp("-",inpfname))
        fn = stdin;
    else
      {
        fn = fopen( inpfname, "r" );
        if (!fn)
          {
            fprintf( stderr, "ERROR: could not open '%s'\n", inpfname );
            return;
          }
        if (debugmode)
          {
#ifdef _WINDOWS
            handle = _fileno( fn );
#else
            handle = fileno( fn );
#endif
            printf( "fopen(\"%s\",\"r\") h=%d\n", inpfname, handle );
          }
      }

    if (!strcmp("-",outfname))
       outf = stdout;
    else
      {
        _unlink( outfname );  /* in case output file is on buggy mars_nfs server */
        outf = fopen( outfname, "w" );
        if (!fn)
          {
            fprintf( stderr, "ERROR: could not create '%s'\n", outfname );
            my_fclose( fn );
            return;
          }
        if (debugmode)
          {
#ifdef _WINDOWS
            handle = _fileno( outf );
#else
            handle = fileno( outf );
#endif
            printf( "fopen(\"%s\",\"w\") h=%d\n", outfname, handle );
          }
      }

    af = init_XML_parse( fn );
    tokn_count = 0;
    for (;;)
      {
        len = XML_fgets( token, MAX_TOKN_LEN, af );
        if (len > MAX_TOKN_LEN)
          {
            fprintf( stderr, "ERR:\a XML_fgets(,%d,)=%d (strlen=%d) on line %ld of %s\n",
                     MAX_TOKN_LEN, len, strlen(token),
                     XML_tokn_line_no, inpfname );
          }
        while (len != (int)strlen(token))
          {
            if (!warned2)
              {
                warned2 = true;
                fprintf( stderr, "ERROR: zero-byte on or near line %ld of %s\n",
                         XML_tokn_line_no, inpfname );
              }
            token[strlen(token)] = ' ';
          }
        if (InpEOF)
            break;
        tokn_count++;

        if (showlinenums)
          {  /* -n */
            if (verboseoutput)
                fprintf( outf, "%ld", XML_tokn_line_no );
            else
                fprintf( outf, "%ld/", XML_tokn_line_no );
          }
        if (verboseoutput)
          {  /* -v */
            if (XML_tokn_is_tag)
                fprintf( outf, "/Tag" );
            else if (XML_tokn_is_comment)
                fprintf( outf, "/Cmt" );
            else
                fprintf( outf, "/Txt" );
            if (XML_tokn_is_truncated)
                fprintf( outf, "/Trunc" );
            if (XML_tokn_is_continuation)
                fprintf( outf, "/Cont" );
            if (!len)
                fprintf( outf, "/NULL\n" );
            else
              {
                if ('\n'==token[len-1])
                  {
                    token[--len] = '\0';
                    fprintf( outf, "+LF" );
                  }
                fprintf( outf, ": '%s'\n", token );
              }
          }
        else
          {
            strcpy( lprefix, "0000:" );
            if (len)
              {
                if (XML_tokn_is_tag)
                    lprefix[0] = '1'; /* tag */
                else if (XML_tokn_is_comment)
                    lprefix[0] = '3'; /* comment */
                else
                    lprefix[0] = '2'; /* text */
                if (XML_tokn_is_truncated)
                    lprefix[1] = '1';
                if (XML_tokn_is_continuation)
                    lprefix[2] = '1';
                if ('\n'==token[len-1])
                  {
                    token[--len] = '\0';
                    lprefix[3] = '1';
                  }
                for (i=0; i<len; i++)
                  { /* no embedded end-line characters permitted in output */
                    if (('\n'==token[i]) || ('\r'==token[i]))
                      {
                        if (!warned)
                          {
                            if (!debugmode)
                                warned = true;  /* only warn about 1st one per file */
                            fprintf( stderr, "ERR:\a embedded char=%d on line %ld of %s\n",
                                     (unsigned int)(unsigned char)(token[i]),
                                     XML_tokn_line_no, inpfname );
                          }
                        token[i] = ' ';
                      }
                  }
              }
            fprintf( outf, "%s%s\n", lprefix, token );
          }
      }/*for*/

    uninit_XML_parse( af );
    af = NULL;
    fn = NULL;

    if (outfname[0])
        my_fclose( outf );

    return;
  } /* lex1file */


#if 0
  #pragma optimize("",off)
#endif

#if 0
  #pragma optimize("",on)
#endif

/*--------------------------------------------------------------------*/


int main( int argc, char *argv[] )
{
 int showhelp = 0;
  while (argv[1])
    {
      if (!strcasecmp(argv[1],"-d")) {
         debugmode = 1;
         if (('=' == argv[1][2]) && argv[1][3]) {
            /* support "-d=32" syntax */
            debugmode = atoi( argv[1]+3 );
         }
      } else if (!strcasecmp(argv[1],"-n")) {
         showlinenums = 1;
      } else if (!strcasecmp(argv[1],"-v")) {
         verboseoutput = 1;
      } else {
         break;
      }
      argc--; argv++;
    }

  infile = "";
  outfile = "";
  if (argc < 2)
      showhelp = 1;  /* erroneous parameters */
  else if (!strcmp("-",argv[1]))
    {
      /* if first parameter is "-" then input is from stdin and everything that follows is the output file name, perhaps w/ blanks in it */
      infile = "-";
      argc--;
      argv++;
      if (1 == argc)
        { /* only file name specified is "-" which means that input is stdin and output is stdout */
          outfile = "-";
        }
      else
        {
          my_strncpy( outfilename, argv[1], sizeof(outfilename)-1 );
          while ((--argc) > 1)
            {
              argv++;
              my_strncat( outfilename, " ", sizeof(outfilename)-1 );
              my_strncat( outfilename, argv[1], sizeof(outfilename)-1 );
            }
          outfile = outfilename;
        }
    }
  else if (!strcmp("-",argv[argc-1]))
    {
      /* if parameters end in "-" then what preceeds it is the input file name, perhaps w/ blanks in it */
      outfile = "-";
      argc--;
      if (1 == argc)
          infile = argv[1];
      else /* argc > 1 */
        {
          my_strncpy( infilename, argv[1], sizeof(infilename)-1 );
          while ((--argc) > 1)
            {
              argv++;
              my_strncat( infilename, " ", sizeof(infilename)-1 );
              my_strncat( infilename, argv[1], sizeof(infilename)-1 );
            }
          infile = infilename;
        }
    }

  if ((!infile[0]) && (argc > 1))
    {
      infile = argv[1];
      argc--;
      argv++;
    }
  if (!outfile[0])
    {
      if (argc > 1)
        {
          outfile = argv[1];
          argc--;
          argv++;
        }
      else
        outfile = "-";
    }

  /* printf( "_MSC_VER=%d\n", _MSC_VER );  --  1200 means MS VC++ 6.0 */
  if (showhelp || (argc > 1) || (!infile[0]))
    {
/*+=>keyflag<=-- "XML_lex v1.%v -- lexical analysis of XML/HTML files.\n"     */
      fprintf( stderr, "XML_lex v1.42 -- lexical analysis of XML/HTML files.\n"
                       "Copyright 2002, Burton Systems Software.  http:\\\\www.burtonsys.com\\\n\n"
                       "Usage:\n"
                       "  XML_lex [options] inpfile [outpfile]\n\n"
                       "If inpfile is \"-\" then input is from stdin.\n"
                       "If outpfile is \"-\" or omitted then output goes to stdout.\n"
                       "If either file is \"-\" then the other may contain blanks, even if not quoted.\n"
                       "Legal options are:\n"
                       "   -v  verbose mode (more human-readable)\n"
                       "   -n  prefix each output line with the input file's line number\n"
                       "   -d  enable debug prints (or -d=63 for verbose debug prints)\n"
                       "Terse output format adds a 5-character prefix (\"abcd:\") to each line:\n"
                       "   a  is 1 for <token>, 2 for text, or 3 for <!--comment-->\n"
                       "   b  is 1 if truncated (continued in next), otherwise 0\n"
                       "   c  is 1 if continued from previous (i.e, previous was truncated), otherwise 0\n"
                       "   d  is 1 if a linefeed/newline character should be appended, otherwise 0\n"
                       "Examples:\n"
                       "   XML_lex inpfile outpfile\n"
                       "   XML_lex inpfile - >outpfile\n"
                       "   XML_lex inpfile | more\n"
                       "   type index.html | XML_lex - outpfile\n" );
      exit(1);
    }

  lex1file( infile, outfile );

  return 0;
}/*main*/



#if 0   /* This is a sample Perl program that uses XML_lex */
/*
if ($#ARGV < 1) {
  give_help:
   print "$0 -- lex and regurgitate HTML or XML file\n" .
         "\n" .
         "Usage:\n" .
         "\n" .
         "   Perl4 $0 {options} inpfile.html outpfile.html\n" .
         "\n" .
         "Example:\n" .
         "   C:\> Perl4 $0 index.html index2.html\n" .
         "   C:\> fc /b index.html index2.html\n" .
         "   Comparing files index.html and index2.html\n" .
         "   FC: no differences encountered\n";

   exit(1);
}

$erlvl = 0;

sub scan_file {
   local($inpfile,$outfile) = @_;
   # print "Scanning $inpfile\n";
   unlink $outfile;
   if (!open(INFILE,"XML_lex $inpfile|")) {
      print STDERR "ERROR: could not pipe from \"XML_lex $inpfile|\", $!\n";
      $erlvl = 1;
   } else {
      if (eof(INFILE)) {
         # print STDERR "ERROR: no input from pipe: \"XML_lex $inpfile|\"\n";
         #  don't need to display error message, because XML_lex already displayed one
         $erlvl = 1;
      } elsif (!open(OUTFILE,">$outfile")) {
         print STDERR "ERROR: could not create $outfile, $!\n";
         $erlvl = 1;
      } else {
         while (<INFILE>) {
            chop;
            $lprefix = substr( $_, 0, 5 );
            $chunk = substr( $_, 5 );
            if ("1" eq substr($lprefix,3,1)) {
               $chunk .= "\n";
            }
            print OUTFILE $chunk;
         }#while
         close OUTFILE;
         $erlvl = 0;
      }
      close INFILE;
   } #else
} #scan_file

&scan_file( $ARGV[0], $ARGV[1] );

exit($erlvl);

__END__
*/
#endif
